import os
import sys
import pickle
import subprocess

from .. import getConfigOption
from ..Qt import QtCore, QtWidgets
from .repl_widget import ReplWidget
from .exception_widget import ExceptionHandlerWidget


class ConsoleWidget(QtWidgets.QWidget):
    """
    Widget displaying console output and accepting command input.
    Implements:
        
      - eval python expressions / exec python statements
      - storable history of commands
      - exception handling allowing commands to be interpreted in the context of any level in the exception stack frame
    
    Why not just use python in an interactive shell (or ipython) ? There are a few reasons:
       
      - pyside does not yet allow Qt event processing and interactive shell at the same time
      - on some systems, typing in the console _blocks_ the qt event loop until the user presses enter. This can
        be baffling and frustrating to users since it would appear the program has frozen.
      - some terminals (eg windows cmd.exe) have notoriously unfriendly interfaces
      - ability to add extra features like exception stack introspection
      - ability to have multiple interactive prompts, including for spawned sub-processes
    """
    def __init__(self, parent=None, namespace=None, historyFile=None, text=None, editor=None):
        """
        ==============  =============================================================================
        **Arguments:**
        namespace       dictionary containing the initial variables present in the default namespace
        historyFile     optional file for storing command history
        text            initial text to display in the console window
        editor          optional string for invoking code editor (called when stack trace entries are 
                        double-clicked). May contain {fileName} and {lineNum} format keys. Example:: 
                      
                            editorCommand --loadfile {fileName} --gotoline {lineNum}
        ==============  =============================================================================
        """
        QtWidgets.QWidget.__init__(self, parent)

        self._setupUi()

        if namespace is None:
            namespace = {}
        namespace['__console__'] = self

        self.localNamespace = namespace
        self.editor = editor
        
        self.output = self.repl.output
        self.input = self.repl.input
        self.input.setFocus()
        
        if text is not None:
            self.output.setPlainText(text)

        self.historyFile = historyFile
        
        try:
            history = self.loadHistory()
        except Exception as exc:
            sys.excepthook(*sys.exc_info())
            history = None
        if history is not None:
            self.input.history = [""] + history
            self.historyList.addItems(history[::-1])

        self.currentTraceback = None

    def _setupUi(self):
        self.layout = QtWidgets.QGridLayout(self)
        self.setLayout(self.layout)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)

        self.splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical, self)
        self.layout.addWidget(self.splitter, 0, 0)

        self.repl = ReplWidget(self.globals, self.locals, self)
        self.splitter.addWidget(self.repl)

        self.historyList = QtWidgets.QListWidget(self)
        self.historyList.hide()
        self.splitter.addWidget(self.historyList)

        self.historyBtn = QtWidgets.QPushButton('History', self)
        self.historyBtn.setCheckable(True)
        self.repl.inputLayout.addWidget(self.historyBtn)

        self.repl.sigCommandEntered.connect(self._commandEntered)
        self.repl.sigCommandRaisedException.connect(self._commandRaisedException)

        self.excHandler = ExceptionHandlerWidget(self)
        self.excHandler.hide()
        self.splitter.addWidget(self.excHandler)

        self.exceptionBtn = QtWidgets.QPushButton("Exceptions..", self)
        self.exceptionBtn.setCheckable(True)
        self.repl.inputLayout.addWidget(self.exceptionBtn)

        self.excHandler.sigStackItemDblClicked.connect(self._stackItemDblClicked)
        self.exceptionBtn.toggled.connect(self.excHandler.setVisible)
        self.historyBtn.toggled.connect(self.historyList.setVisible)
        self.historyList.itemClicked.connect(self.cmdSelected)
        self.historyList.itemDoubleClicked.connect(self.cmdDblClicked)

    def catchAllExceptions(self, catch=True):
        if catch:
            self.exceptionBtn.setChecked(True)
        self.excHandler.catchAllExceptions(catch)

    def catchNextException(self, catch=True):
        if catch:
            self.exceptionBtn.setChecked(True)
        self.excHandler.catchNextException(catch)

    def setStack(self, frame=None):
        self.excHandler.setStack(frame)
 
    def loadHistory(self):
        """Return the list of previously-invoked command strings (or None)."""
        if self.historyFile is not None and os.path.exists(self.historyFile):
            with open(self.historyFile, 'rb') as pf:
                return pickle.load(pf)
        
    def saveHistory(self, history):
        """Store the list of previously-invoked command strings."""
        if self.historyFile is not None:
            with open(self.historyFile, 'wb') as pf:
                pickle.dump(history, pf)
        
    def _commandEntered(self, repl, cmd):
        self.historyList.addItem(cmd)
        self.saveHistory(self.input.history[1:100])
        sb = self.historyList.verticalScrollBar()
        sb.setValue(sb.maximum())

    def _commandRaisedException(self, repl, exc):
        self.excHandler.exceptionHandler(exc)

    def globals(self):
        frame = self.excHandler.selectedFrame()
        if frame is not None and self.excHandler.runSelectedFrameCheck.isChecked():
            return frame.f_globals
        else:
            return self.localNamespace
        
    def locals(self):
        frame = self.excHandler.selectedFrame()
        if frame is not None and self.excHandler.runSelectedFrameCheck.isChecked():
            return frame.f_locals
        else:
            return self.localNamespace

    def cmdSelected(self, item):
        index = -(self.historyList.row(item)+1)
        self.input.setHistory(index)
        self.input.setFocus()
        
    def cmdDblClicked(self, item):
        index = -(self.historyList.row(item)+1)
        self.input.setHistory(index)
        self.input.execCmd()
        
    def _stackItemDblClicked(self, handler, item):
        editor = self.editor
        if editor is None:
            editor = getConfigOption('editorCommand')
        if editor is None:
            return
        tb = self.excHandler.selectedFrame()
        lineNum = tb.f_lineno
        fileName = tb.f_code.co_filename
        subprocess.Popen(self.editor.format(fileName=fileName, lineNum=lineNum), shell=True)



import ctypes
import itertools
import numpy as np
from . import QT_LIB, QtCore, QtGui
from . import compat

__all__ = ["get_qpainterpath_element_array"]

if QT_LIB.startswith('PyQt'):
    from . import sip
elif QT_LIB == 'PySide2':
    from PySide2 import __version_info__ as pyside_version_info
elif QT_LIB == 'PySide6':
    from PySide6 import __version_info__ as pyside_version_info

class QArrayDataQt5(ctypes.Structure):
    _fields_ = [
        ("ref", ctypes.c_int),
        ("size", ctypes.c_int),
        ("alloc", ctypes.c_uint, 31),
        ("offset", ctypes.c_ssize_t),
    ]

class QPainterPathPrivateQt5(ctypes.Structure):
    _fields_ = [
        ("ref", ctypes.c_int),
        ("adata", ctypes.POINTER(QArrayDataQt5)),
    ]

class QArrayDataQt6(ctypes.Structure):
    _fields_ = [
        ("ref", ctypes.c_int),
        ("flags", ctypes.c_uint),
        ("alloc", ctypes.c_ssize_t),
    ]

class QPainterPathPrivateQt6(ctypes.Structure):
    _fields_ = [
        ("ref", ctypes.c_int),
        ("adata", ctypes.POINTER(QArrayDataQt6)),
        ("data", ctypes.c_void_p),
        ("size", ctypes.c_ssize_t),
    ]

def get_qpainterpath_element_array(qpath, nelems=None):
    writable = nelems is not None
    if writable:
        qpath.reserve(nelems)

    itemsize = 24
    dtype = dict(names=['x','y','c'], formats=['f8', 'f8', 'i4'], itemsize=itemsize)

    ptr0 = compat.unwrapinstance(qpath)
    pte_cp = ctypes.c_void_p.from_address(ptr0)
    if not pte_cp:
        return np.zeros(0, dtype=dtype)

    # _cp : ctypes pointer
    # _ci : ctypes instance
    if QT_LIB in ['PyQt5', 'PySide2']:
        PTR = ctypes.POINTER(QPainterPathPrivateQt5)
        pte_ci = ctypes.cast(pte_cp, PTR).contents
        size_ci = pte_ci.adata[0]
        eptr = ctypes.addressof(size_ci) + size_ci.offset
    elif QT_LIB in ['PyQt6', 'PySide6']:
        PTR = ctypes.POINTER(QPainterPathPrivateQt6)
        pte_ci = ctypes.cast(pte_cp, PTR).contents
        size_ci = pte_ci
        eptr = pte_ci.data
    else:
        raise NotImplementedError

    if writable:
        size_ci.size = nelems
    else:
        nelems = size_ci.size

    vp = compat.voidptr(eptr, itemsize*nelems, writable)
    return np.frombuffer(vp, dtype=dtype)

class PrimitiveArray:
    # Note: This class is an internal implementation detail and is not part
    #       of the public API.
    #
    # QPainter has a C++ native API that takes an array of objects:
    #   drawPrimitives(const Primitive *array, int count, ...)
    # where "Primitive" is one of QPointF, QLineF, QRectF, PixmapFragment
    #
    # PySide (with the exception of drawPixmapFragments) and older PyQt
    # require a Python list of "Primitive" instances to be provided to
    # the respective "drawPrimitives" method.
    #
    # This is inefficient because:
    # 1) constructing the Python list involves calling wrapinstance multiple times.
    #    - this is mitigated here by reusing the instance pointers
    # 2) The binding will anyway have to repack the instances into a contiguous array,
    #    in order to call the underlying C++ native API.
    #
    # Newer PyQt provides sip.array, which is more efficient.
    #
    # PySide's drawPixmapFragments() takes an instance to the first item of a
    # C array of PixmapFragment(s) _and_ the length of the array.
    # There is no overload that takes a Python list of PixmapFragment(s).

    def __init__(self, Klass, nfields, *, use_array=None):
        self._Klass = Klass
        self._nfields = nfields
        self._capa = -1

        self.use_sip_array = False
        self.use_ptr_to_array = False

        if QT_LIB.startswith('PyQt'):
            if use_array is None:
                use_array = (
                    hasattr(sip, 'array') and
                    (
                        (0x60301 <= QtCore.PYQT_VERSION) or
                        (0x50f07 <= QtCore.PYQT_VERSION < 0x60000)
                    )
                )
            self.use_sip_array = use_array

        if QT_LIB.startswith('PySide'):
            if use_array is None:
                use_array = (
                    Klass is QtGui.QPainter.PixmapFragment
                    or pyside_version_info >= (6, 4, 3)
                )
            self.use_ptr_to_array = use_array

        self.resize(0)

    def resize(self, size):
        if self.use_sip_array:
            # For reference, SIP_VERSION 6.7.8 first arrived
            # in PyQt5_sip 12.11.2 and PyQt6_sip 13.4.2
            if sip.SIP_VERSION >= 0x60708:
                if size <= self._capa:
                    self._size = size
                    return
            else:
                # sip.array prior to SIP_VERSION 6.7.8 had a
                # buggy slicing implementation.
                # so trigger a reallocate for any different size
                if size == self._capa:
                    return

            self._siparray = sip.array(self._Klass, size)

        else:
            if size <= self._capa:
                self._size = size
                return
            self._ndarray = np.empty((size, self._nfields), dtype=np.float64)

            if self.use_ptr_to_array:
                # defer creation
                self._objs = None
            else:
                self._objs = self._wrap_instances(self._ndarray)

        self._capa = size
        self._size = size

    def _wrap_instances(self, array):
        return list(map(compat.wrapinstance,
            itertools.count(array.ctypes.data, array.strides[0]),
            itertools.repeat(self._Klass, array.shape[0])))

    def __len__(self):
        return self._size

    def ndarray(self):
        # ndarray views are cheap to recreate each time
        if self.use_sip_array:
            if sip.SIP_VERSION >= 0x60708:
                mv = self._siparray
            else:
                # sip.array prior to SIP_VERSION 6.7.8 had a buggy buffer protocol
                # that set the wrong size.
                # workaround it by going through a sip.voidptr
                mv = sip.voidptr(self._siparray, self._capa*self._nfields*8)
            # note that we perform the slicing by using only _size rows
            nd = np.frombuffer(mv, dtype=np.float64, count=self._size*self._nfields)
            return nd.reshape((-1, self._nfields))
        else:
            return self._ndarray[:self._size]

    def instances(self):
        # this returns an iterable container of Klass instances.
        # for "use_ptr_to_array" mode, such a container may not
        # be required at all, so its creation is deferred
        if self.use_sip_array:
            if self._size == self._capa:
                # avoiding slicing when it's not necessary
                # handles the case where sip.array had a buggy
                # slicing implementation 
                return self._siparray
            else:
                # this is a view
                return self._siparray[:self._size]

        if self._objs is None:
            self._objs = self._wrap_instances(self._ndarray)

        if self._size == self._capa:
            return self._objs
        else:
            # this is a shallow copy
            return self._objs[:self._size]

    def drawargs(self):
        # returns arguments to apply to the respective drawPrimitives() functions
        if self.use_ptr_to_array:
            if self._capa > 0:
                # wrap memory only if it is safe to do so
                ptr = compat.wrapinstance(self._ndarray.ctypes.data, self._Klass)
            else:
                # shiboken translates None <--> nullptr
                # alternatively, we could instantiate a dummy _Klass()
                ptr = None
            return ptr, self._size

        else:
            return self.instances(),



import traceback
import numpy as np

import pyqtgraph as pg
from pyqtgraph.graphicsItems.ScatterPlotItem import name_list
from pyqtgraph.Qt import QtWidgets, QtCore
from pyqtgraph.parametertree import interact, ParameterTree, Parameter
import random

pg.mkQApp()

rng = np.random.default_rng(10)
random.seed(10)


def sortedRandint(low, high, size):
    return np.sort(rng.integers(low, high, size))


def isNoneOrScalar(value):
    return value is None or np.isscalar(value[0])


values = {
    # Convention 1
    "None (replaced by integer indices)": None,
    # Convention 2
    "Single curve values": sortedRandint(0, 20, 15),
    # Convention 3 list form
    "container of (optionally) mixed-size curve values": [
        sortedRandint(0, 20, 15),
        *[sortedRandint(0, 20, 15) for _ in range(4)],
    ],
    # Convention 3 array form
    "2D matrix": np.row_stack([sortedRandint(20, 40, 15) for _ in range(6)]),
}


def next_plot(xtype="random", ytype="random", symbol="o", symbolBrush="#f00"):
    constKwargs = locals()
    x = y = None
    if xtype == "random":
        xtype = random.choice(list(values))
    if ytype == "random":
        ytype = random.choice(list(values))
    x = values[xtype]
    y = values[ytype]
    textbox.setValue(f"x={xtype}\ny={ytype}")
    pltItem.clear()
    try:
        pltItem.multiDataPlot(
            x=x, y=y, pen=cmap.getLookupTable(nPts=6), constKwargs=constKwargs
        )
    except Exception as e:
        QtWidgets.QMessageBox.critical(widget, "Error", traceback.format_exc())


cmap = pg.colormap.get("viridis")
widget = pg.PlotWidget()
pltItem: pg.PlotItem = widget.plotItem

xytype = dict(type="list", values=list(values))
topParam = interact(
    next_plot,
    symbolBrush=dict(type="color"),
    symbol=dict(type="list", values=name_list),
    xtype=xytype,
    ytype=xytype,
)
tree = ParameterTree()
tree.setMinimumWidth(150)
tree.addParameters(topParam, showTop=True)

textbox = Parameter.create(name="text", type="text", readonly=True)
tree.addParameters(textbox)

win = QtWidgets.QWidget()
win.setLayout(lay := QtWidgets.QHBoxLayout())
lay.addWidget(widget)
lay.addWidget(tree)
if __name__ == "__main__":
    win.show()
    pg.exec()



from .. import functions as fn
from .. import getConfigOption
from ..Point import Point
from ..Qt import QtCore, QtWidgets
from .GraphicsObject import GraphicsObject
from .GraphicsWidgetAnchor import GraphicsWidgetAnchor
from .TextItem import TextItem

__all__ = ['ScaleBar']

class ScaleBar(GraphicsWidgetAnchor, GraphicsObject):
    """
    Displays a rectangular bar to indicate the relative scale of objects on the view.
    """
    def __init__(self, size, width=5, brush=None, pen=None, suffix='m', offset=None):
        GraphicsObject.__init__(self)
        GraphicsWidgetAnchor.__init__(self)
        self.setFlag(self.GraphicsItemFlag.ItemHasNoContents)
        self.setAcceptedMouseButtons(QtCore.Qt.MouseButton.NoButton)
        
        if brush is None:
            brush = getConfigOption('foreground')
        self.brush = fn.mkBrush(brush)
        self.pen = fn.mkPen(pen)
        self._width = width
        self.size = size
        if offset is None:
            offset = (0,0)
        self.offset = offset
        
        self.bar = QtWidgets.QGraphicsRectItem()
        self.bar.setPen(self.pen)
        self.bar.setBrush(self.brush)
        self.bar.setParentItem(self)
        
        self.text = TextItem(text=fn.siFormat(size, suffix=suffix), anchor=(0.5,1))
        self.text.setParentItem(self)

    def changeParent(self):
        view = self.parentItem()
        if view is None:
            return
        view.sigRangeChanged.connect(self.updateBar)
        self.updateBar()
        
        
    def updateBar(self):
        view = self.parentItem()
        if view is None:
            return
        p1 = view.mapFromViewToItem(self, QtCore.QPointF(0,0))
        p2 = view.mapFromViewToItem(self, QtCore.QPointF(self.size,0))
        w = (p2-p1).x()
        self.bar.setRect(QtCore.QRectF(-w, 0, w, self._width))
        self.text.setPos(-w/2., 0)

    def boundingRect(self):
        return QtCore.QRectF()

    def setParentItem(self, p):
        ret = GraphicsObject.setParentItem(self, p)
        if self.offset is not None:
            offset = Point(self.offset)
            anchorx = 1 if offset[0] <= 0 else 0
            anchory = 1 if offset[1] <= 0 else 0
            anchor = (anchorx, anchory)
            self.anchor(itemPos=anchor, parentPos=anchor, offset=offset)
        return ret



import sys, traceback
from ..Qt import QtWidgets, QtGui


class StackWidget(QtWidgets.QTreeWidget):
    def __init__(self, parent=None):
        QtWidgets.QTreeWidget.__init__(self, parent)
        self.setAlternatingRowColors(True)
        self.setHeaderHidden(True)

    def selectedFrame(self):
        """Return the currently selected stack frame (or None if there is no selection)
        """
        sel = self.selectedItems()
        if len(sel) == 0:
            return None
        else:
            return sel[0].frame

    def clear(self):
        QtWidgets.QTreeWidget.clear(self)
        self.frames = []

    def setException(self, exc=None, lastFrame=None):
        """Display an exception chain with its tracebacks and call stack.
        """
        if exc is None:
            exc = sys.exc_info()[1]

        self.clear()

        exceptions = exceptionChain(exc)
        for ex, cause in exceptions:
            stackFrames, tbFrames = stacksFromTraceback(ex.__traceback__, lastFrame=lastFrame)
            catchMsg = textItem("Exception caught here")
            excStr = ''.join(traceback.format_exception_only(type(ex), ex)).strip()
            items = makeItemTree(stackFrames + [catchMsg] + tbFrames, excStr)
            self.addTopLevelItem(items[0])
            if cause is not None:
                if cause == 'cause':
                    causeItem = textItem("The above exception was the direct cause of the following exception:")
                elif cause == 'context':
                    causeItem = textItem("During handling of the above exception, another exception occurred:")
                self.addTopLevelItem(causeItem)

        items[0].setExpanded(True)

    def setStack(self, frame=None, expand=True, lastFrame=None):
        """Display a call stack and exception traceback.

        This allows the user to probe the contents of any frame in the given stack.

        *frame* may either be a Frame instance or None, in which case the current 
        frame is retrieved from ``sys._getframe()``. 

        If *tb* is provided then the frames in the traceback will be appended to 
        the end of the stack list. If *tb* is None, then sys.exc_info() will 
        be checked instead.
        """
        if frame is None:
            frame = sys._getframe().f_back

        self.clear()

        stack = stackFromFrame(frame, lastFrame=lastFrame)
        items = makeItemTree(stack, "Call stack")
        self.addTopLevelItem(items[0])
        if expand:
            items[0].setExpanded(True)

    
def stackFromFrame(frame, lastFrame=None):
    """Return (text, stack_frame) for the entire stack ending at *frame*

    If *lastFrame* is given and present in the stack, then the stack is truncated 
    at that frame.
    """
    lines = traceback.format_stack(frame)
    frames = []
    while frame is not None:
        frames.insert(0, frame)
        frame = frame.f_back
    if lastFrame is not None and lastFrame in frames:
        frames = frames[:frames.index(lastFrame)+1]
        
    return list(zip(lines[:len(frames)], frames))


def stacksFromTraceback(tb, lastFrame=None):
    """Return (text, stack_frame) for a traceback and the stack preceding it

    If *lastFrame* is given and present in the stack, then the stack is truncated 
    at that frame.
    """
    # get stack before tb
    stack = stackFromFrame(tb.tb_frame.f_back if tb is not None else lastFrame)
    if tb is None:
        return stack, []

    # walk to last frame of traceback        
    lines = traceback.format_tb(tb)
    frames = []
    while True:            
        frames.append(tb.tb_frame)
        if tb.tb_next is None or tb.tb_frame is lastFrame:
            break
        tb = tb.tb_next

    return stack, list(zip(lines[:len(frames)], frames))


def makeItemTree(stack, title):
    topItem = QtWidgets.QTreeWidgetItem([title])
    topItem.frame = None
    font = topItem.font(0)
    font.setWeight(font.Weight.Bold)
    topItem.setFont(0, font)
    items = [topItem]
    for entry in stack:
        if isinstance(entry, QtWidgets.QTreeWidgetItem):
            item = entry
        else:
            text, frame = entry
            item = QtWidgets.QTreeWidgetItem([text.rstrip()])
            item.frame = frame
        topItem.addChild(item)
        items.append(item)
    return items


def exceptionChain(exc):
    """Return a list of (exception, 'cause'|'context') pairs for exceptions
    leading up to *exc*
    """
    exceptions = [(exc, None)]
    while True:
        # walk through chained exceptions
        if exc.__cause__ is not None:
            exc = exc.__cause__
            exceptions.insert(0, (exc, 'cause'))
        elif exc.__context__ is not None and exc.__suppress_context__ is False:
            exc = exc.__context__
            exceptions.insert(0, (exc, 'context'))
        else:
            break
    return exceptions


def textItem(text):
    """Return a tree item with no associated stack frame and a darker background color
    """
    item = QtWidgets.QTreeWidgetItem([text])
    item.frame = None
    item.setBackground(0, QtGui.QBrush(QtGui.QColor(220, 220, 220)))
    item.setForeground(0, QtGui.QBrush(QtGui.QColor(0, 0, 0)))
    item.setChildIndicatorPolicy(item.ChildIndicatorPolicy.DontShowIndicator)
    return item



import sys, re, traceback, threading
from .. import exceptionHandling as exceptionHandling
from ..Qt import QtWidgets, QtCore
from ..functions import SignalBlock
from .stackwidget import StackWidget


class ExceptionHandlerWidget(QtWidgets.QGroupBox):
    sigStackItemClicked = QtCore.Signal(object, object)  # self, item
    sigStackItemDblClicked = QtCore.Signal(object, object)  # self, item
    _threadException = QtCore.Signal(object)
    
    def __init__(self, parent=None):
        super().__init__(parent)
        self._setupUi()

        self.filterString = ''
        self._inSystrace = False

        # send exceptions raised in non-gui threads back to the main thread by signal.
        self._threadException.connect(self._threadExceptionHandler)

    def _setupUi(self):
        self.setTitle("Exception Handling")

        self.layout = QtWidgets.QGridLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setHorizontalSpacing(2)
        self.layout.setVerticalSpacing(0)

        self.clearExceptionBtn = QtWidgets.QPushButton("Clear Stack", self)
        self.clearExceptionBtn.setEnabled(False)
        self.layout.addWidget(self.clearExceptionBtn, 0, 6, 1, 1)

        self.catchAllExceptionsBtn = QtWidgets.QPushButton("Show All Exceptions", self)
        self.catchAllExceptionsBtn.setCheckable(True)
        self.layout.addWidget(self.catchAllExceptionsBtn, 0, 1, 1, 1)

        self.catchNextExceptionBtn = QtWidgets.QPushButton("Show Next Exception", self)
        self.catchNextExceptionBtn.setCheckable(True)
        self.layout.addWidget(self.catchNextExceptionBtn, 0, 0, 1, 1)

        self.onlyUncaughtCheck = QtWidgets.QCheckBox("Only Uncaught Exceptions", self)
        self.onlyUncaughtCheck.setChecked(True)
        self.layout.addWidget(self.onlyUncaughtCheck, 0, 4, 1, 1)

        self.stackTree = StackWidget(self)
        self.layout.addWidget(self.stackTree, 2, 0, 1, 7)

        self.runSelectedFrameCheck = QtWidgets.QCheckBox("Run commands in selected stack frame", self)
        self.runSelectedFrameCheck.setChecked(True)
        self.layout.addWidget(self.runSelectedFrameCheck, 3, 0, 1, 7)

        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.layout.addItem(spacerItem, 0, 5, 1, 1)

        self.filterLabel = QtWidgets.QLabel("Filter (regex):", self)
        self.layout.addWidget(self.filterLabel, 0, 2, 1, 1)

        self.filterText = QtWidgets.QLineEdit(self)
        self.layout.addWidget(self.filterText, 0, 3, 1, 1)

        self.catchAllExceptionsBtn.toggled.connect(self.catchAllExceptions)
        self.catchNextExceptionBtn.toggled.connect(self.catchNextException)
        self.clearExceptionBtn.clicked.connect(self.clearExceptionClicked)
        self.stackTree.itemClicked.connect(self.stackItemClicked)
        self.stackTree.itemDoubleClicked.connect(self.stackItemDblClicked)
        self.onlyUncaughtCheck.toggled.connect(self.updateSysTrace)
        self.filterText.textChanged.connect(self._filterTextChanged)

    def setStack(self, frame=None):
        self.clearExceptionBtn.setEnabled(True)
        self.stackTree.setStack(frame)

    def setException(self, exc=None, lastFrame=None):
        self.clearExceptionBtn.setEnabled(True)
        self.stackTree.setException(exc, lastFrame=lastFrame)

    def selectedFrame(self):
        return self.stackTree.selectedFrame()

    def catchAllExceptions(self, catch=True):
        """
        If True, the console will catch all unhandled exceptions and display the stack
        trace. Each exception caught clears the last.
        """
        with SignalBlock(self.catchAllExceptionsBtn.toggled, self.catchAllExceptions):
            self.catchAllExceptionsBtn.setChecked(catch)
        
        if catch:
            with SignalBlock(self.catchNextExceptionBtn.toggled, self.catchNextException):
                self.catchNextExceptionBtn.setChecked(False)
            self.enableExceptionHandling()
        else:
            self.disableExceptionHandling()
        
    def catchNextException(self, catch=True):
        """
        If True, the console will catch the next unhandled exception and display the stack
        trace.
        """
        with SignalBlock(self.catchNextExceptionBtn.toggled, self.catchNextException):
            self.catchNextExceptionBtn.setChecked(catch)
        if catch:
            with SignalBlock(self.catchAllExceptionsBtn.toggled, self.catchAllExceptions):
                self.catchAllExceptionsBtn.setChecked(False)
            self.enableExceptionHandling()
        else:
            self.disableExceptionHandling()
        
    def enableExceptionHandling(self):
        exceptionHandling.registerCallback(self.exceptionHandler)
        self.updateSysTrace()
        
    def disableExceptionHandling(self):
        exceptionHandling.unregisterCallback(self.exceptionHandler)
        self.updateSysTrace()
        
    def clearExceptionClicked(self):
        self.stackTree.clear()
        self.clearExceptionBtn.setEnabled(False)
        
    def updateSysTrace(self):
        ## Install or uninstall  sys.settrace handler 
        
        if not self.catchNextExceptionBtn.isChecked() and not self.catchAllExceptionsBtn.isChecked():
            if sys.gettrace() == self.systrace:
                self._disableSysTrace()
            return
        
        if self.onlyUncaughtCheck.isChecked():
            if sys.gettrace() == self.systrace:
                self._disableSysTrace()
        else:
            if sys.gettrace() not in (None, self.systrace):
                self.onlyUncaughtCheck.setChecked(False)
                raise Exception("sys.settrace is in use (are you using another debugger?); cannot monitor for caught exceptions.")
            else:
                self._enableSysTrace()

    def _enableSysTrace(self):
        # set global trace function
        # note: this has no effect on pre-existing frames or threads 
        # until settrace_all_threads arrives in python 3.12.
        sys.settrace(self.systrace)  # affects current thread only
        threading.settrace(self.systrace)  # affects new threads only
        if hasattr(threading, 'settrace_all_threads'):
            threading.settrace_all_threads(self.systrace)

    def _disableSysTrace(self):
        sys.settrace(None)
        threading.settrace(None)
        if hasattr(threading, 'settrace_all_threads'):
            threading.settrace_all_threads(None)

    def exceptionHandler(self, excInfo, lastFrame=None):
        if isinstance(excInfo, Exception):
            exc = excInfo
        else:
            exc = excInfo.exc_value

        # exceptions raised in non-gui threads must be sent to the gui thread by signal
        isGuiThread = QtCore.QThread.currentThread() == QtCore.QCoreApplication.instance().thread()
        if not isGuiThread:
            # note: we are giving the user the ability to modify a frame owned by another thread.. 
            # expect trouble :)
            self._threadException.emit((excInfo, lastFrame))
            return

        if self.catchNextExceptionBtn.isChecked():
            self.catchNextExceptionBtn.setChecked(False)
        elif not self.catchAllExceptionsBtn.isChecked():
            return
        
        self.setException(exc, lastFrame=lastFrame)
    
    def _threadExceptionHandler(self, args):
        self.exceptionHandler(*args)

    def systrace(self, frame, event, arg):
        if event != 'exception':
            return self.systrace

        if self._inSystrace:
            # prevent recursve calling
            return self.systrace
        self._inSystrace = True
        try:
            if self.checkException(*arg):
                # note: the exception has no __traceback__ at this point!
                self.exceptionHandler(arg[1], lastFrame=frame)
        except Exception as exc:
            print("Exception in systrace:")
            traceback.print_exc()
        finally:
            self.inSystrace = False
        return self.systrace
        
    def checkException(self, excType, exc, tb):
        ## Return True if the exception is interesting; False if it should be ignored.
        
        filename = tb.tb_frame.f_code.co_filename
        function = tb.tb_frame.f_code.co_name
        
        filterStr = self.filterString
        if filterStr != '':
            if isinstance(exc, Exception):
                msg = traceback.format_exception_only(type(exc), exc)
            elif isinstance(exc, str):
                msg = exc
            else:
                msg = repr(exc)
            match = re.search(filterStr, "%s:%s:%s" % (filename, function, msg))
            return match is not None

        ## Go through a list of common exception points we like to ignore:
        if excType is GeneratorExit or excType is StopIteration:
            return False
        if excType is AttributeError:
            if filename.endswith('numpy/core/fromnumeric.py') and function in ('all', '_wrapit', 'transpose', 'sum'):
                return False
            if filename.endswith('numpy/core/arrayprint.py') and function in ('_array2string'):
                return False
            if filename.endswith('MetaArray.py') and function == '__getattr__':
                for name in ('__array_interface__', '__array_struct__', '__array__'):  ## numpy looks for these when converting objects to array
                    if name in exc:
                        return False
            if filename.endswith('flowchart/eq.py'):
                return False
        if excType is TypeError:
            if filename.endswith('numpy/lib/function_base.py') and function == 'iterable':
                return False
            
        return True
    
    def stackItemClicked(self, item):
        self.sigStackItemClicked.emit(self, item)

    def stackItemDblClicked(self, item):
        self.sigStackItemDblClicked.emit(self, item)

    def _filterTextChanged(self, value):
        self.filterString = str(value)



"""
Display a plot and an image with minimal setup. 

pg.plot() and pg.image() are indended to be used from an interactive prompt
to allow easy data inspection (but note that PySide unfortunately does not
call the Qt event loop while the interactive prompt is running, in this case
it is necessary to call QApplication.exec_() to make the windows appear).
"""

import numpy as np

import pyqtgraph as pg

data = np.random.normal(size=1000)
pg.plot(data, title="Simplest possible plotting example")

data = np.random.normal(size=(500,500))
pg.image(data, title="Simplest possible image example")

if __name__ == '__main__':
    pg.exec()



from math import atan2, degrees

import numpy as np

from .Point import Point
from .Qt import QtGui
from . import SRTTransform3D


class SRTTransform(QtGui.QTransform):
    """Transform that can always be represented as a combination of 3 matrices: scale * rotate * translate
    This transform has no shear; angles are always preserved.
    """
    def __init__(self, init=None):
        QtGui.QTransform.__init__(self)
        self.reset()
        
        if init is None:
            return
        elif isinstance(init, dict):
            self.restoreState(init)
        elif isinstance(init, SRTTransform):
            self._state = {
                'pos': Point(init._state['pos']),
                'scale': Point(init._state['scale']),
                'angle': init._state['angle']
            }
            self.update()
        elif isinstance(init, QtGui.QTransform):
            self.setFromQTransform(init)
        elif isinstance(init, QtGui.QMatrix4x4):
            self.setFromMatrix4x4(init)
        else:
            raise Exception("Cannot create SRTTransform from input type: %s" % str(type(init)))

        
    def getScale(self):
        return self._state['scale']
        
    def getRotation(self):
        return self._state['angle']
        
    def getTranslation(self):
        return self._state['pos']
    
    def reset(self):
        self._state = {
            'pos': Point(0,0),
            'scale': Point(1,1),
            'angle': 0.0  ## in degrees
        }
        self.update()
        
    def setFromQTransform(self, tr):
        p1 = Point(tr.map(0., 0.))
        p2 = Point(tr.map(1., 0.))
        p3 = Point(tr.map(0., 1.))
        
        dp2 = Point(p2-p1)
        dp3 = Point(p3-p1)
        
        ## detect flipped axes
        if dp2.angle(dp3, units="radians") > 0:
            da = 0
            sy = -1.0
        else:
            da = 0
            sy = 1.0
            
        self._state = {
            'pos': Point(p1),
            'scale': Point(dp2.length(), dp3.length() * sy),
            'angle': degrees(atan2(dp2[1], dp2[0])) + da
        }
        self.update()
        
    def setFromMatrix4x4(self, m):
        m = SRTTransform3D.SRTTransform3D(m)
        angle, axis = m.getRotation()
        if angle != 0 and (axis[0] != 0 or axis[1] != 0 or axis[2] != 1):
            print("angle: %s  axis: %s" % (str(angle), str(axis)))
            raise Exception("Can only convert 4x4 matrix to 3x3 if rotation is around Z-axis.")
        self._state = {
            'pos': Point(m.getTranslation()),
            'scale': Point(m.getScale()),
            'angle': angle
        }
        self.update()
        
    def translate(self, *args):
        """Acceptable arguments are: 
           x, y
           [x, y]
           Point(x,y)"""
        t = Point(*args)
        self.setTranslate(self._state['pos']+t)
        
    def setTranslate(self, *args):
        """Acceptable arguments are: 
           x, y
           [x, y]
           Point(x,y)"""
        self._state['pos'] = Point(*args)
        self.update()
        
    def scale(self, *args):
        """Acceptable arguments are: 
           x, y
           [x, y]
           Point(x,y)"""
        s = Point(*args)
        self.setScale(self._state['scale'] * s)
        
    def setScale(self, *args):
        """Acceptable arguments are: 
           x, y
           [x, y]
           Point(x,y)"""
        self._state['scale'] = Point(*args)
        self.update()
        
    def rotate(self, angle):
        """Rotate the transformation by angle (in degrees)"""
        self.setRotate(self._state['angle'] + angle)
        
    def setRotate(self, angle):
        """Set the transformation rotation to angle (in degrees)"""
        self._state['angle'] = angle
        self.update()

    def __truediv__(self, t):
        """A / B  ==  B^-1 * A"""
        dt = t.inverted()[0] * self
        return SRTTransform(dt)
        
    def __div__(self, t):
        return self.__truediv__(t)
        
    def __mul__(self, t):
        return SRTTransform(QtGui.QTransform.__mul__(self, t))

    def saveState(self):
        p = self._state['pos']
        s = self._state['scale']
        #if s[0] == 0:
            #raise Exception('Invalid scale: %s' % str(s))
        return {'pos': (p[0], p[1]), 'scale': (s[0], s[1]), 'angle': self._state['angle']}

    def restoreState(self, state):
        self._state['pos'] = Point(state.get('pos', (0,0)))
        self._state['scale'] = Point(state.get('scale', (1.,1.)))
        self._state['angle'] = state.get('angle', 0)
        self.update()

    def update(self):
        QtGui.QTransform.reset(self)
        ## modifications to the transform are multiplied on the right, so we need to reverse order here.
        QtGui.QTransform.translate(self, *self._state['pos'])
        QtGui.QTransform.rotate(self, self._state['angle'])
        QtGui.QTransform.scale(self, *self._state['scale'])

    def __repr__(self):
        return str(self.saveState())
        
    def matrix(self):
        return np.array([[self.m11(), self.m12(), self.m13()],[self.m21(), self.m22(), self.m23()],[self.m31(), self.m32(), self.m33()]])



"""
Using ConsoleWidget to interactively inspect exception backtraces


TODO
 - fix uncaught exceptions in threads (python 3.12)
 - allow using qtconsole
 - provide thread info for stacks
 - add thread browser?
 - add object browser?
    - clicking on a stack frame populates list of locals?
 - optional merged exception stacks

"""

import sys
import queue
import functools
import threading
import pyqtgraph as pg
import pyqtgraph.console
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.debug import threadName


def raiseException():
    """Raise an exception
    """
    x = "inside raiseException()"
    raise Exception(f"Raised an exception {x} in {threadName()}")


def raiseNested():
    """Raise an exception while handling another
    """
    x = "inside raiseNested()"
    try:
        raiseException()
    except Exception:
        raise Exception(f"Raised during exception handling {x} in {threadName()}")


def raiseFrom():
    """Raise an exception from another
    """
    x = "inside raiseFrom()"
    try:
        raiseException()
    except Exception as exc:
        raise Exception(f"Raised-from during exception handling {x} in {threadName()}") from exc


def raiseCaughtException():
    """Raise and catch an exception
    """
    x = "inside raiseCaughtException()"
    try:
        raise Exception(f"Raised an exception {x} in {threadName()}")
    except Exception:
        print(f"Raised and caught exception {x} in {threadName()}  trace: {sys._getframe().f_trace}")


def captureStack():
    """Inspect the curent call stack
    """
    x = "inside captureStack()"
    global console
    console.setStack()
    return x


# Background thread for running functions
threadRunQueue = queue.Queue()
def threadRunner():
    global threadRunQueue
    # This is necessary to allow installing trace functions in the thread later on
    sys.settrace(lambda *args: None)
    while True:
        func, args = threadRunQueue.get()
        try:
            print(f"running {func} from thread, trace: {sys._getframe().f_trace}")
            func(*args)
        except Exception:
            sys.excepthook(*sys.exc_info())
thread = threading.Thread(target=threadRunner, name="background_thread", daemon=True)
thread.start()


# functions used to generate a stack a few items deep
def runInStack(func):
    x = "inside runInStack(func)"
    runInStack2(func)
    return x

def runInStack2(func):
    x = "inside runInStack2(func)"
    runInStack3(func)
    return x

def runInStack3(func):
    x = "inside runInStack3(func)"
    runInStack4(func)
    return x

def runInStack4(func):
    x = "inside runInStack4(func)"
    func()
    return x


class SignalEmitter(pg.QtCore.QObject):
    signal = pg.QtCore.Signal(object, object)
    def __init__(self, queued):
        pg.QtCore.QObject.__init__(self)
        if queued:
            self.signal.connect(self.run, pg.QtCore.Qt.ConnectionType.QueuedConnection)
        else:
            self.signal.connect(self.run)
    def run(self, func, args):
        func(*args)
signalEmitter = SignalEmitter(queued=False)
queuedSignalEmitter = SignalEmitter(queued=True)



def runFunc(func):
    if signalCheck.isChecked():
        if queuedSignalCheck.isChecked():
            func = functools.partial(queuedSignalEmitter.signal.emit, runInStack, (func,))
        else:
            func = functools.partial(signalEmitter.signal.emit, runInStack, (func,))
    
    if threadCheck.isChecked():
        threadRunQueue.put((runInStack, (func,)))
    else:
        runInStack(func)



funcs = [
    raiseException,
    raiseNested,
    raiseFrom,
    raiseCaughtException,
    captureStack,
]

app = pg.mkQApp()

win = pg.QtWidgets.QSplitter(pg.QtCore.Qt.Orientation.Horizontal)

ctrl = QtWidgets.QWidget()
ctrlLayout = QtWidgets.QVBoxLayout()
ctrl.setLayout(ctrlLayout)
win.addWidget(ctrl)

btns = []
for func in funcs:
    btn = QtWidgets.QPushButton(func.__doc__)
    btn.clicked.connect(functools.partial(runFunc, func))
    btns.append(btn)
    ctrlLayout.addWidget(btn)

threadCheck = QtWidgets.QCheckBox('Run in thread')
ctrlLayout.addWidget(threadCheck)

signalCheck = QtWidgets.QCheckBox('Run from Qt signal')
ctrlLayout.addWidget(signalCheck)

queuedSignalCheck = QtWidgets.QCheckBox('Use queued Qt signal')
ctrlLayout.addWidget(queuedSignalCheck)

ctrlLayout.addStretch()

console = pyqtgraph.console.ConsoleWidget(text="""
Use ConsoleWidget to interactively inspect exception tracebacks and call stacks!

- Enable "Show next exception" and the next unhandled exception will be displayed below.
- Click any of the buttons to the left to generate an exception.
- When an exception traceback is shown, you can select any of the stack frames and then run commands from that context,
  allowing you to inspect variables along the stack. (hint: most of the functions called by the buttons to the left 
  have a variable named "x" in their local scope)
- Note that this is not like a typical debugger--the program is not paused when an exception is caught; we simply keep
  a reference to the stack frames and continue on.
- By default, we only catch unhandled exceptions. If you need to inspect a handled exception (one that is caught by
  a try:except block), then uncheck the "Only handled exceptions" box. Note, however that this incurs a performance 
  penalty and will interfere with other debuggers.


""")
console.catchNextException()
win.addWidget(console)

win.resize(1400, 800)
win.setSizes([300, 1100])
win.show()

if __name__ == '__main__':
    pg.exec()



import warnings

from ...Qt import QtCore
from .action import ParameterControlledButton
from .basetypes import GroupParameter, GroupParameterItem
from ..ParameterItem import ParameterItem
from ...Qt import QtCore, QtWidgets


class ActionGroupParameterItem(GroupParameterItem):
    """
    Wraps a :class:`GroupParameterItem` to manage function parameters created by
    an interactor. Provies a button widget which mimics an ``action`` parameter.
    """

    def __init__(self, param, depth):
        self.itemWidget = QtWidgets.QWidget()
        self.button = ParameterControlledButton(parent=self.itemWidget)
        self.button.clicked.connect(param.activate)

        self.itemWidget.setLayout(layout := QtWidgets.QHBoxLayout())
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.button)

        super().__init__(param, depth)

    def treeWidgetChanged(self):
        ParameterItem.treeWidgetChanged(self)
        tw = self.treeWidget()
        if tw is None:
            return
        tw.setItemWidget(self, 1, self.itemWidget)

    def optsChanged(self, param, opts):
        if "button" in opts:
            buttonOpts = opts["button"] or dict(visible=False)
            self.button.updateOpts(param, buttonOpts)
        super().optsChanged(param, opts)


class ActionGroupParameter(GroupParameter):
    itemClass = ActionGroupParameterItem

    sigActivated = QtCore.Signal(object)

    def __init__(self, **opts):
        opts.setdefault("button", {})
        super().__init__(**opts)

    def activate(self):
        self.sigActivated.emit(self)
        self.emitStateChanged('activated', None)

    def setButtonOpts(self, **opts):
        """
        Update individual button options without replacing the entire
        button definition.
        """
        buttonOpts = self.opts.get("button", {}).copy()
        buttonOpts.update(opts)
        self.setOpts(button=buttonOpts)


class ActionGroup(ActionGroupParameter):
    sigActivated = QtCore.Signal()

    def __init__(self, **opts):
        warnings.warn(
            "`ActionGroup` is deprecated and will be removed in the first release after "
            "January 2023. Use `ActionGroupParameter` instead. See "
            "https://github.com/pyqtgraph/pyqtgraph/pull/2505 for details.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(**opts)

    def activate(self):
        self.sigActivated.emit()



import numpy as np

from .. import functions as fn
from .. import getConfigOption
from .. import Qt
from ..Qt import QtCore, QtGui
from .GraphicsObject import GraphicsObject

__all__ = ['BarGraphItem']

class BarGraphItem(GraphicsObject):
    def __init__(self, **opts):
        """
        Valid keyword options are:
        x, x0, x1, y, y0, y1, width, height, pen, brush
        
        x specifies the x-position of the center of the bar.
        x0, x1 specify left and right edges of the bar, respectively.
        width specifies distance from x0 to x1.
        You may specify any combination:
            
            x, width
            x0, width
            x1, width
            x0, x1
            
        Likewise y, y0, y1, and height. 
        If only height is specified, then y0 will be set to 0
        
        Example uses:
        
            BarGraphItem(x=range(5), height=[1,5,2,4,3], width=0.5)
            
        
        """
        GraphicsObject.__init__(self)
        self.opts = dict(
            x=None,
            y=None,
            x0=None,
            y0=None,
            x1=None,
            y1=None,
            name=None,
            height=None,
            width=None,
            pen=None,
            brush=None,
            pens=None,
            brushes=None,
        )

        if 'pen' not in opts:
            opts['pen'] = getConfigOption('foreground')
        if 'brush' not in opts:
            opts['brush'] = (128, 128, 128)
        # the first call to _updateColors() will thus always be an update

        self._rectarray = Qt.internals.PrimitiveArray(QtCore.QRectF, 4)
        self._shape = None
        self.picture = None
        self.setOpts(**opts)
        
    def setOpts(self, **opts):
        self.opts.update(opts)
        self.picture = None
        self._shape = None
        self._prepareData()
        self._updateColors(opts)
        self.prepareGeometryChange()
        self.update()
        self.informViewBoundsChanged()

    def _updatePenWidth(self, pen):
        no_pen = pen is None or pen.style() == QtCore.Qt.PenStyle.NoPen
        if no_pen:
            return

        idx = pen.isCosmetic()
        self._penWidth[idx] = max(self._penWidth[idx], pen.widthF())

    def _updateColors(self, opts):
        # the logic here is to permit the user to update only data
        # without updating pens/brushes

        # update only if fresh pen/pens supplied
        if 'pen' in opts or 'pens' in opts:
            self._penWidth = [0, 0]

            if self.opts['pens'] is None:
                # pens not configured, use single pen
                pen = self.opts['pen']
                pen = fn.mkPen(pen)
                self._updatePenWidth(pen)
                self._sharedPen = pen
                self._pens = None
            else:
                # pens configured, ignore single pen (if any)
                pens = []
                for pen in self.opts['pens']:
                    if not isinstance(pen, QtGui.QPen):
                        pen = fn.mkPen(pen)
                    pens.append(pen)
                self._updatePenWidth(pen)
                self._sharedPen = None
                self._pens = pens

        # update only if fresh brush/brushes supplied
        if 'brush' in opts or 'brushes' in opts:
            if self.opts['brushes'] is None:
                # brushes not configured, use single brush
                brush = self.opts['brush']
                self._sharedBrush = fn.mkBrush(brush)
                self._brushes = None
            else:
                # brushes configured, ignore single brush (if any)
                brushes = []
                for brush in self.opts['brushes']:
                    if not isinstance(brush, QtGui.QBrush):
                        brush = fn.mkBrush(brush)
                    brushes.append(brush)
                self._sharedBrush = None
                self._brushes = brushes

        self._singleColor = (
            self._sharedPen is not None and
            self._sharedBrush is not None
        )

    def _getNormalizedCoords(self):
        def asarray(x):
            if x is None or np.isscalar(x) or isinstance(x, np.ndarray):
                return x
            return np.array(x)

        x = asarray(self.opts.get('x'))
        x0 = asarray(self.opts.get('x0'))
        x1 = asarray(self.opts.get('x1'))
        width = asarray(self.opts.get('width'))
        
        if x0 is None:
            if width is None:
                raise Exception('must specify either x0 or width')
            if x1 is not None:
                x0 = x1 - width
            elif x is not None:
                x0 = x - width/2.
            else:
                raise Exception('must specify at least one of x, x0, or x1')
        if width is None:
            if x1 is None:
                raise Exception('must specify either x1 or width')
            width = x1 - x0
            
        y = asarray(self.opts.get('y'))
        y0 = asarray(self.opts.get('y0'))
        y1 = asarray(self.opts.get('y1'))
        height = asarray(self.opts.get('height'))

        if y0 is None:
            if height is None:
                y0 = 0
            elif y1 is not None:
                y0 = y1 - height
            elif y is not None:
                y0 = y - height/2.
            else:
                y0 = 0
        if height is None:
            if y1 is None:
                raise Exception('must specify either y1 or height')
            height = y1 - y0

        # ensure x0 < x1 and y0 < y1
        t0, t1 = x0, x0 + width
        x0 = np.minimum(t0, t1, dtype=np.float64)
        x1 = np.maximum(t0, t1, dtype=np.float64)
        t0, t1 = y0, y0 + height
        y0 = np.minimum(t0, t1, dtype=np.float64)
        y1 = np.maximum(t0, t1, dtype=np.float64)

        # here, all of x0, y0, x1, y1 are numpy objects,
        # BUT could possibly be numpy scalars
        return x0, y0, x1, y1

    def _prepareData(self):
        x0, y0, x1, y1 = self._getNormalizedCoords()
        if x0.size == 0 or y0.size == 0:
            self._dataBounds = (None, None), (None, None)
            self._rectarray.resize(0)
            return

        xmn, xmx = np.min(x0), np.max(x1)
        ymn, ymx = np.min(y0), np.max(y1)
        self._dataBounds = (xmn, xmx), (ymn, ymx)

        self._rectarray.resize(max(x0.size, y0.size))
        memory = self._rectarray.ndarray()
        memory[:, 0] = x0
        memory[:, 1] = y0
        memory[:, 2] = x1 - x0
        memory[:, 3] = y1 - y0

    def _render(self, painter):
        multi_pen = self._pens is not None
        multi_brush = self._brushes is not None
        no_pen = (
            not multi_pen
            and self._sharedPen.style() == QtCore.Qt.PenStyle.NoPen
        )

        rects = self._rectarray.instances()

        if no_pen and multi_brush:
            for idx, rect in enumerate(rects):
                painter.fillRect(rect, self._brushes[idx])
        else:
            if not multi_pen:
                painter.setPen(self._sharedPen)
            if not multi_brush:
                painter.setBrush(self._sharedBrush)

            for idx, rect in enumerate(rects):
                if multi_pen:
                    painter.setPen(self._pens[idx])
                if multi_brush:
                    painter.setBrush(self._brushes[idx])

                painter.drawRect(rect)

    def drawPicture(self):
        self.picture = QtGui.QPicture()
        painter = QtGui.QPainter(self.picture)
        self._render(painter)
        painter.end()

    def paint(self, p, *args):
        if self._singleColor:
            p.setPen(self._sharedPen)
            p.setBrush(self._sharedBrush)
            drawargs = self._rectarray.drawargs()
            p.drawRects(*drawargs)
        else:
            if self.picture is None:
                self.drawPicture()
            self.picture.play(p)
            
    def shape(self):
        if self._shape is None:
            shape = QtGui.QPainterPath()
            rects = self._rectarray.instances()
            for rect in rects:
                shape.addRect(rect)
            self._shape = shape
        return self._shape

    def implements(self, interface=None):
        ints = ['plotData']
        if interface is None:
            return ints
        return interface in ints

    def name(self):
        return self.opts.get('name', None)

    def getData(self):
        return self.opts.get('x'),  self.opts.get('height')

    def dataBounds(self, ax, frac=1.0, orthoRange=None):
        # _penWidth is available after _updateColors()
        pw = self._penWidth[0] * 0.5
        # _dataBounds is available after _prepareData()
        bounds = self._dataBounds[ax]
        if bounds[0] is None or bounds[1] is None:
            return None, None

        return (bounds[0] - pw, bounds[1] + pw)

    def pixelPadding(self):
        # _penWidth is available after _updateColors()
        pw = (self._penWidth[1] or 1) * 0.5
        return pw

    def boundingRect(self):
        xmn, xmx = self.dataBounds(ax=0)
        if xmn is None or xmx is None:
            return QtCore.QRectF()
        ymn, ymx = self.dataBounds(ax=1)
        if ymn is None or ymx is None:
            return QtCore.QRectF()

        px = py = 0
        pxPad = self.pixelPadding()
        if pxPad > 0:
            # determine length of pixel in local x, y directions
            px, py = self.pixelVectors()
            px = 0 if px is None else px.length()
            py = 0 if py is None else py.length()
            # return bounds expanded by pixel size
            px *= pxPad
            py *= pxPad

        return QtCore.QRectF(xmn-px, ymn-py, (2*px)+xmx-xmn, (2*py)+ymx-ymn)



import code, sys, traceback
from ..Qt import QtWidgets, QtGui, QtCore
from ..functions import mkBrush
from .CmdInput import CmdInput


class ReplWidget(QtWidgets.QWidget):
    sigCommandEntered = QtCore.Signal(object, object)  # self, command
    sigCommandRaisedException = QtCore.Signal(object, object)  # self, exc

    def __init__(self, globals, locals, parent=None):
        self.globals = globals
        self.locals = locals
        self._lastCommandRow = None
        self._commandBuffer = []  # buffer to hold multiple lines of input
        self.stdoutInterceptor = StdoutInterceptor(self.write)
        self.ps1 = ">>> "
        self.ps2 = "... "

        QtWidgets.QWidget.__init__(self, parent=parent)

        self._setupUi()

        # define text styles
        isDark = self.output.palette().color(QtGui.QPalette.ColorRole.Base).value() < 128
        outputBlockFormat = QtGui.QTextBlockFormat()
        outputFirstLineBlockFormat = QtGui.QTextBlockFormat(outputBlockFormat)
        outputFirstLineBlockFormat.setTopMargin(5)
        outputCharFormat = QtGui.QTextCharFormat()
        outputCharFormat.setFontWeight(QtGui.QFont.Weight.Normal)
        cmdBlockFormat = QtGui.QTextBlockFormat()
        cmdBlockFormat.setBackground(mkBrush("#335" if isDark else "#CCF"))
        cmdCharFormat = QtGui.QTextCharFormat()
        cmdCharFormat.setFontWeight(QtGui.QFont.Weight.Bold)
        self.textStyles = {
            'command': (cmdCharFormat, cmdBlockFormat),
            'output': (outputCharFormat, outputBlockFormat),
            'output_first_line': (outputCharFormat, outputFirstLineBlockFormat),
        }

        self.input.ps1 = self.ps1
        self.input.ps2 = self.ps2

    def _setupUi(self):
        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.setLayout(self.layout)

        self.output = QtWidgets.QTextEdit(self)
        font = QtGui.QFont()
        font.setFamily("Courier New")
        font.setStyleStrategy(QtGui.QFont.StyleStrategy.PreferAntialias)
        self.output.setFont(font)
        self.output.setReadOnly(True)
        self.layout.addWidget(self.output)
        
        # put input box in a horizontal layout so we can easily place buttons at the end
        self.inputWidget = QtWidgets.QWidget(self)
        self.layout.addWidget(self.inputWidget)
        self.inputLayout = QtWidgets.QHBoxLayout()
        self.inputWidget.setLayout(self.inputLayout)
        self.inputLayout.setContentsMargins(0, 0, 0, 0)

        self.input = CmdInput(parent=self)
        self.inputLayout.addWidget(self.input)

        self.input.sigExecuteCmd.connect(self.runCmd)

    def runCmd(self, cmd):
        if '\n' in cmd:
            for line in cmd.split('\n'):
                self.runCmd(line)
            return

        if len(self._commandBuffer) == 0:
            self.write(f"{self.ps1}{cmd}\n", style='command')
        else:
            self.write(f"{self.ps2}{cmd}\n", style='command')
        
        self.sigCommandEntered.emit(self, cmd)
        self._commandBuffer.append(cmd)

        fullcmd = '\n'.join(self._commandBuffer)
        try:
            cmdCode = code.compile_command(fullcmd)
            self.input.setMultiline(False)
        except Exception:
            # cannot continue processing this command; reset and print exception
            self._commandBuffer = []
            self.displayException()
            self.input.setMultiline(False)
        else:
            if cmdCode is None:
                # incomplete input; wait for next line
                self.input.setMultiline(True)
                return

            self._commandBuffer = []

            # run command
            try:
                with self.stdoutInterceptor:
                    exec(cmdCode, self.globals(), self.locals())
            except Exception as exc:
                self.displayException()
                self.sigCommandRaisedException.emit(self, exc)

            # Add a newline if the output did not
            cursor = self.output.textCursor()
            if cursor.columnNumber() > 0:
                self.write('\n')

    def write(self, strn, style='output', scrollToBottom='auto'):
        """Write a string into the console.

        If scrollToBottom is 'auto', then the console is automatically scrolled
        to fit the new text only if it was already at the bottom.
        """
        isGuiThread = QtCore.QThread.currentThread() == QtCore.QCoreApplication.instance().thread()
        if not isGuiThread:
            sys.__stdout__.write(strn)
            return

        cursor = self.output.textCursor()
        cursor.movePosition(QtGui.QTextCursor.MoveOperation.End)
        self.output.setTextCursor(cursor)

        sb = self.output.verticalScrollBar()
        scroll = sb.value()
        if scrollToBottom == 'auto':
            atBottom = scroll == sb.maximum()
            scrollToBottom = atBottom

        row = cursor.blockNumber()
        if style == 'command':
            self._lastCommandRow = row

        if style == 'output' and row == self._lastCommandRow + 1:
            # adjust style for first line of output
            firstLine, endl, strn = strn.partition('\n')
            self._setTextStyle('output_first_line')
            self.output.insertPlainText(firstLine + endl)

        if len(strn) > 0:
            self._setTextStyle(style)
            self.output.insertPlainText(strn)
            # return to output style immediately to avoid seeing an extra line of command style
            if style != 'output':
                self._setTextStyle('output')

        if scrollToBottom:
            sb.setValue(sb.maximum())
        else:
            sb.setValue(scroll)

    def displayException(self):
        """
        Display the current exception and stack.
        """
        tb = traceback.format_exc()
        lines = []
        indent = 4
        prefix = '' 
        for l in tb.split('\n'):
            lines.append(" "*indent + prefix + l)
        self.write('\n'.join(lines))
        
    def _setTextStyle(self, style):
        charFormat, blockFormat = self.textStyles[style]
        cursor = self.output.textCursor()
        cursor.setBlockFormat(blockFormat)
        self.output.setCurrentCharFormat(charFormat)


class StdoutInterceptor:
    """Used to temporarily redirect writes meant for sys.stdout and sys.stderr to a new location
    """
    def __init__(self, writeFn):
        self._orig_stdout = None
        self._orig_stderr = None
        self.writeFn = writeFn

    def realOutputFiles(self):
        """Return the real sys.stdout and stderr (which are sometimes masked while running commands)
        """
        return (
            self._orig_stdout or sys.stdout,
            self._orig_stderr or sys.stderr
        )

    def print(self, *args):
        """Print to real stdout (for debugging)
        """
        self.realOutputFiles()[0].write(' '.join(map(str, args)) + "\n")

    def flush(self):
        # Need to implement this since we temporarily occlude sys.stdout
        pass

    def fileno(self):
        # Need to implement this since we temporarily occlude sys.stdout, and someone may be looking for it (faulthandler, for example)
        return 1

    def write(self, strn):
        self.writeFn(strn)

    def __enter__(self):
        self._orig_stdout = sys.stdout
        self._orig_stderr = sys.stderr
        sys.stdout = self
        sys.stderr = self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self._orig_stdout
        sys.stderr = self._orig_stderr
        self._orig_stdout = None
        self._orig_stderr = None

